#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <math.h>

#include <grass/gis.h>

#define PROGVERSION 1.1

int
main (int argc, char *argv[])
{
	struct GModule *module;
	
	struct
	{
		struct Option *input;
		struct Option *output;
		struct Option *radius;
	} parm;
	
	struct
	{
		struct Flag *rect;
		struct Flag *absolute;
		struct Flag *quiet;
	} flag;
			
	char *sysstr;
	char *mapset;
	int i, j, radius, n, x, y;
	int a, b, dist;
	int firstrun;
	int nullvalue;
	double centerval;
	double sum, prominence;
	double min, max;
	double from, to;
	int nrows, ncols;
	int fd, fd_out;
	struct Cell_head region;
	struct Colors colors;
	DCELL *diskrow = NULL;
	DCELL *outrow = NULL;
		
	sysstr = G_calloc (512, sizeof (char));
	module = G_define_module ();
	module->description = "Calculates prominence statistic";
	/* setup some basic GIS stuff */
	G_gisinit (argv[0]);
	/* do not pause after a warning message was displayed */
	G_sleep_on_error (0);
		
	parm.input = G_define_standard_option(G_OPT_R_INPUT);
	parm.input->required = YES;
	parm.input->description = "Raster map for which to calculate index";
	
	parm.output = G_define_standard_option(G_OPT_R_OUTPUT);
	parm.output->required = YES;
	parm.output->description = "Raster to store output";	
	
	parm.radius = G_define_option ();
	parm.radius->key = "radius";
	parm.radius->type = TYPE_INTEGER;
	parm.radius->required = YES;
	parm.radius->description = "Radius of neighborhood in map cells";

	flag.absolute = G_define_flag ();
	flag.absolute->key = 'a';
	flag.absolute->description = "Calculate absolute differences";
	
	flag.rect = G_define_flag ();
	flag.rect->key = 'r';
	flag.rect->description = "Square neighborhood shape (instead of circular)";
		
	flag.quiet = G_define_flag ();
	flag.quiet->key = 'q';
	flag.quiet->description = "Disable on-screen progress display";

		
	/* parse command line */
	if (G_parser (argc, argv))
	{
		exit (-1);
	}

	G_get_window (&region);
	nrows = G_window_rows ();
	ncols = G_window_cols ();
	
	/* check parameters for validity */
	radius = atoi (parm.radius->answer);
	if (( radius < 1 ) || ( radius > nrows/2 ) || ( radius > ncols/2 )) {
		G_fatal_error ("Radius must be > 1 and smaller than half the map's extent in both directions.");
	}
	mapset = G_calloc (512, sizeof (char));
	mapset = G_find_cell (parm.input->answer, "");
	if ( mapset == NULL) {
		G_fatal_error ("Input map does not exist in the current location.");
	}
	fd = G_open_cell_old (parm.input->answer, mapset);	
	if (fd < 0) {
		G_fatal_error ("Could not open input map for reading!\n");
	}	
	if (!G_legal_filename (parm.output->answer)) {
		G_fatal_error ("%s is not a legal filename for an output map.\n",
				parm.output->answer);
	}
	fd_out = G_open_raster_new (parm.output->answer, DCELL_TYPE);
	if ( fd_out < 0 ) {
		G_fatal_error ("Could not open output map for writing!\n");
	}
	
	diskrow = G_allocate_d_raster_buf ();
	outrow = G_allocate_d_raster_buf ();	
	
	/* initialize statistics */
	n = 0;
	firstrun = 1;
	min = 0;
	max = 0;
	sum = 0.0;
	
	if (!flag.quiet->answer) {
		fprintf (stdout,"Calculating prominence index:\n");
		fflush (stdout);
	}
	
	/* main loop: read raster row from disk, add to values in neighborhood */
	i = 0;
	j = 0;
	nullvalue = 0;
	for (y = 0; y < nrows; y ++) {
		for (x = 0; x < ncols; x++) {
			G_get_d_raster_row (fd, diskrow, y);
			centerval = diskrow [x];
			if (G_is_d_null_value ( &diskrow[x] )) {
				nullvalue = 1;
			} else {
				nullvalue = 0;
				for (j = y-radius; j <= y+radius; j++) {
					if ((j>0) && (j<nrows)) {
						G_get_d_raster_row (fd, diskrow, j);
						nullvalue = 0;
						for (i = x-radius; i <= x+radius; i++) {
							if ( ( (i!=x) || (j!=y)  )  && (i>0) && (i<ncols)) {
								if (!G_is_d_null_value (&diskrow[i])) {
									if (flag.rect->answer) {
										/* use rectangular neighborhood */
										n = n + 1;
										if (flag.absolute->answer) {					
											sum = sum + fabs ( centerval - diskrow [i]);
										} else {
											sum = sum + ( centerval - diskrow [i]);
										}
									} else {
										/* use circular neighborhood */
										a = abs ( x - i );
										b = abs ( y - j );
										dist = (int) sqrt ( (double) (a*a) + (b*b) );
										if (dist <= radius) {
											n = n + 1;
											if (flag.absolute->answer) {
												sum = sum + fabs ( centerval - diskrow [i]);
											} else {
												sum = sum + ( centerval - diskrow [i]);
											}
										}
									}
								}
							}	
						}
					}
				}
			}
			/* write one cell to output map */
			prominence = sum/n;
			if (!nullvalue) {
				outrow [x] = prominence;
				if (firstrun) {
					min = prominence;
					max = prominence;
					firstrun = 0;
				} else {
					if ( prominence < min ) min = prominence;
					if ( prominence > max ) max = prominence;
				}
				n = 0;
				sum = 0.0;
			} else {
				G_set_d_null_value ( &outrow[x],1 );
			}
		}
		/* write one row to output map on disk */
		G_put_raster_row (fd_out, outrow, DCELL_TYPE);
		
		if (!flag.quiet->answer) {
			G_percent (y, nrows-1, 1);
			fflush (stdout);
		}
	}

	G_close_cell (fd);
	G_close_cell (fd_out);

	
	G_init_colors (&colors);

	fprintf (stderr, "SCALE FROM: %f TO: %f\n", min, max);

	if  ( (flag.absolute->answer) ) {
		/* write a colormap in grey shades: lowest prominence gets black, highest white */
		from = min;
		to = max/2;
		G_add_d_raster_color_rule (&from, 0, 0, 0, &to, 127, 127, 127, &colors);
		from = max/2;
		to = max;
		G_add_d_raster_color_rule (&from, 128, 128, 128, &to, 255,  255, 255, &colors);
	} else {
		/* write a colormap from red (-1) thru black (0) to white (1) */
		if ( min < 0 ) {
			from = min;
			to = min/2;
			G_add_d_raster_color_rule (&from, 255, 0, 0, &to, 127,  0, 0, &colors);
			from = min/2;
			to = 0;
			G_add_d_raster_color_rule (&from, 127, 0, 0, &to, 0,  0, 0, &colors);
			from = 0;
		}
		else {
			from = min;
		}
		to = max/2;
		G_add_d_raster_color_rule (&from, 0, 0, 0, &to, 127,  127, 127, &colors);
		from = max/2;
		to = max;
		G_add_d_raster_color_rule (&from, 128, 128, 128, &to, 255,  255, 255, &colors);		
	}
	G_write_colors (parm.output->answer, G_mapset(), &colors);
		
	G_free_colors (&colors);	
	G_free (diskrow);
	G_free (outrow);
	free (mapset);

	if (!flag.quiet->answer) {
		fprintf (stdout,"\nDone.\n");
		fflush (stdout);
	}
				
	return (EXIT_SUCCESS);
}
